// Test app for tracer.dll.
//
// This app demonstrates the use of the Tracer class to wrap/unwrap
// a dynamic sink around an object (either at the client site or the
// object site) to providing automatic tracing of calls to methods.
//
// When Tracer.Wrap(f) is called in Main, the sink is placed in front of the proxy
// (if any).  In this case, you'll see "cli-True" in the tracer output,
// indicating that the sink is on the client side of the connection.
//
// When Tracer.Wrap(this) is called in the Foo constructor, the sink is placed in
// front of the object in whatever context the object resides in.  In this
// case, you'll see "cli-False" in the tracer output, indicating that the
// sink is on the object side of the connection.
//
// Mike Woodring
// http://staff.develop.com/woodring
//

using System;
using System.Runtime.Remoting;
using System.Runtime.Remoting.Contexts;

// Apply the following Synchronization attribute to test with Foo in a context other than
// the default context.  Remove the attribute to test with the Foo instance in
// the default context.
//
[Synchronization(SynchronizationAttribute.REQUIRED)]
class Foo : ContextBoundObject
{
    public Foo()
    {
        // Trace method invocations at the object site (i.e., in the context
        // where the object lives).
        //
        Tracer.Wrap(this, "InFoo");
        n = 0;
    }

    public void Bar()
    {
        Console.WriteLine("Foo.Bar, tid = {0}, ctx={1}", Tracer.GetCurrentThreadId(), Tracer.GetCurrentContextId());
    }

    public int Bar( ref int x, ref double y )
    {
        Console.WriteLine( "Foo.Bar, tid = {0}, ctx={1}, x = {2}, y = {3}",
                            Tracer.GetCurrentThreadId(), Tracer.GetCurrentContextId(),
                            x, y );
        x = -x;
        y = -y;

        return(x + (int)y);
    }

    public int Bar( String s )
    {
        Console.WriteLine("Foo.Bar, tid = {0}, ctx={1}", Tracer.GetCurrentThreadId(), Tracer.GetCurrentContextId());
        return(++n);
    }

    public void Bar( String inString, out String outString )
    {
        Console.WriteLine("Foo.Bar, tid = {0}, ctx={1}", Tracer.GetCurrentThreadId(), Tracer.GetCurrentContextId());
        outString = "You passed: " + inString;
    }

    private int n;
}

//[System.Runtime.Remoting.ThreadAffinity(System.Runtime.Remoting.Synchronization.REQUIRES_NEW)]
public class Test
{
    static public int Main( System.String[] Args )
    {
        Test t = new Test();
        return t._Main(Args);
    }

    int _Main( System.String[] Args )
    {
        Foo f = new Foo();

        // Trace calls to Foo at the client site (here in the default context).
        //
        Tracer.Wrap(f, "MainWrap");
        
        // Or try these variations:
        //
        // Tracer.Wrap(f, "MainWrap", "trace.out");
        // Tracer.Wrap(f, "MainWrap", System.Console.OpenStandardOutput());

        for( int n = 0; n < 3; n++ )
        {
            Console.WriteLine( "\r\nCalling wrapped Foo.Bar, tid = {0}, ctx = {1}",
                               Tracer.GetCurrentThreadId(), Tracer.GetCurrentContextId() );

            int x = 42;
            double y = 3.1415927;

            f.Bar(ref x, ref y);
        }

        // Test unwrapping - should now only see the output from the object-site
        // sink.
        Tracer.Unwrap(f);

        for( int n = 0; n < 3; n++ )
        {
            Console.WriteLine( "\r\nCalling unwrapped Foo.Bar, tid = {0}, ctx = {1}",
                               Tracer.GetCurrentThreadId(), Tracer.GetCurrentContextId() );

            int x = 42;
            double y = 3.1415927;

            f.Bar(ref x, ref y);
        }

        return(0);
    }
}
