// Tracer
//
// The Tracer class allows a dynamic sink to be "wrapped" around an object
// reference to any ContextBoundObject.  Calling Tracer.Wrap places the
// sink in front of the object reference that's passed in.  If the object
// reference passed in is a proxy, the sink is installed at the client
// side.  If the object reference passed in is a direct object reference,
// the sink is installed at the object site.
//
// Clients can use:
//
//      SomeType o = new SomeType(args);
//      Tracer.Wrap(o, "description")
//
// to wrap all calls to a specific object from the client site (if o is a proxy).
// If o happens to be a direct reference, then the client site is the same as
// the object site.  Later, the client can use:
//
//      Tracer.Unwrap(o)
//
// to remove the interceptor.
//
// If you want to trace all calls to a specific *type*, then the
// constructor for that type can use:
//
//      public SomeType()
//      {
//          Tracer.Wrap(this, "SomeType");
//          ...
//      }
//
// to install a sink at the object-site.  This results in an interceptor
// being placed in front of each instance of SomeType.
//
// The text argument to Wrap is just descriptive text that can be included
// in the output.
//
// Mike Woodring
// http://staff.develop.com/woodring
//

using System;
using System.Runtime.Remoting;
using System.Runtime.Remoting.Contexts;
using System.Runtime.Remoting.Messaging;
using System.Runtime.Remoting.Activation;
using System.Collections;
using System.IO;

class TracerSink : IDynamicMessageSink
{
    private String              Description;
    private Tracer.OutputFlags  Flags;
    private StreamWriter        OutStream;

    public TracerSink( String s, Tracer.OutputFlags Flags, StreamWriter OutStream )
    {
        Description = s;
        this.Flags = Flags;
        this.OutStream = OutStream;
    }

    public void ProcessMessageStart( IMessage reqMsg,  bool bClientSide, bool bAsync )
    {
        TraceMethod(reqMsg, "enter", bClientSide, bAsync);
    }

    public void ProcessMessageFinish( IMessage replyMsg, bool bClientSide, bool bAsync )
    {
        TraceMethod(replyMsg, "return", bClientSide, bAsync);
    }

    void TraceMethod( IMessage msg, String note, bool bClientSide, bool bAsync )
    {
        if( msg is IMethodCallMessage )
        {
            TraceMethodEntry(msg as IMethodCallMessage, bClientSide, bAsync);
        }
        else if( msg is IMethodReturnMessage )
        {
            TraceMethodReturn(msg as IMethodReturnMessage, bClientSide, bAsync);
        }
        else
        {
            throw new Exception("*** unexpected message type ***");
        }
    }

    bool IsFlagSet( Tracer.OutputFlags Flag )
    {
        return((Flags & Flag) == Flag ? true : false);
    }

    void TraceMethodEntry( IMethodCallMessage msg, bool bClientSide, bool bAsync )
    {
        if( !IsFlagSet(Tracer.OutputFlags.MethodEntry) )
        {
            return;
        }

        System.Text.StringBuilder TraceText = new System.Text.StringBuilder();

        if( IsFlagSet(Tracer.OutputFlags.Description) )
        {
            TraceText.AppendFormat("[entry, {0},", Description);
        }
        else
        {
            TraceText.Append("[entry,");
        }

        if( IsFlagSet(Tracer.OutputFlags.Date) )
        {
            TraceText.AppendFormat( " {0}/{1}/{2}",
                                    System.DateTime.Today.Month,
                                    System.DateTime.Today.Day,
                                    System.DateTime.Today.Year );
        }

        if( IsFlagSet(Tracer.OutputFlags.Time) )
        {
            TraceText.AppendFormat(" {0}", System.DateTime.Now.TimeOfDay);
        }
        else if( IsFlagSet(Tracer.OutputFlags.TimeMS) )
        {
            TraceText.AppendFormat(" {0}ms", System.DateTime.Now.Millisecond);
        }
        else if( IsFlagSet(Tracer.OutputFlags.TimeTicks) )
        {
            TraceText.AppendFormat(" tick-{0}", System.DateTime.Now.Ticks);
        }

        if( IsFlagSet(Tracer.OutputFlags.PID) )
        {
            TraceText.AppendFormat(" pid-{0}", Tracer.GetCurrentProcessId());
        }

        if( IsFlagSet(Tracer.OutputFlags.TID) )
        {
            TraceText.AppendFormat(" tid-{0}", Tracer.GetCurrentThreadId());
        }

        if( IsFlagSet(Tracer.OutputFlags.Domain) )
        {
            TraceText.AppendFormat(" ad-{0}", Tracer.GetCurrentAppDomainName());
        }

        if( IsFlagSet(Tracer.OutputFlags.Context) )
        {
            TraceText.AppendFormat(" ctx-{0}", Tracer.GetCurrentContextId());
        }

        if( IsFlagSet(Tracer.OutputFlags.SinkSite) )
        {
            TraceText.AppendFormat(" {0}", (bClientSide ? "at-client" : "at-object"));
        }

        if( IsFlagSet(Tracer.OutputFlags.CallNature) )
        {
            TraceText.AppendFormat(" async-{0}", bAsync.ToString());
        }

        TraceText.AppendFormat( "]\r\n  {0}.{1}(",
                                msg.TypeName, msg.MethodName );

        Object[] args = (Object[])msg.MethodSignature;

        if( args.Length > 0 )
        {
            for( int n = 0; n < args.Length; n++ )
            {
                TraceText.AppendFormat( "{0}{1} {2} ({3})",
                                        (n == 0 ? " " : ", "),
                                        args[n].ToString(),
                                        msg.GetArgName(n),
                                        msg.GetArg(n) );
            }

            TraceText.Append(" )");
        }
        else
        {
            TraceText.Append(" void )");
        }

        OutStream.WriteLine(TraceText);
        OutStream.Flush();
    }

    void TraceMethodReturn( IMethodReturnMessage msg, bool bClientSide, bool bAsync )
    {
        if( !IsFlagSet(Tracer.OutputFlags.MethodExit) )
        {
            return;
        }

        System.Text.StringBuilder TraceText = new System.Text.StringBuilder();

        if( IsFlagSet(Tracer.OutputFlags.Description) )
        {
            TraceText.AppendFormat("[exit, {0},", Description);
        }
        else
        {
            TraceText.Append("[exit,");
        }

        if( IsFlagSet(Tracer.OutputFlags.Date) )
        {
            TraceText.AppendFormat( " {0}/{1}/{2}",
                                    System.DateTime.Today.Month,
                                    System.DateTime.Today.Day,
                                    System.DateTime.Today.Year );
        }

        if( IsFlagSet(Tracer.OutputFlags.Time) )
        {
            TraceText.AppendFormat(" {0}", System.DateTime.Now.TimeOfDay);
        }
        else if( IsFlagSet(Tracer.OutputFlags.TimeMS) )
        {
            TraceText.AppendFormat(" {0}ms", System.DateTime.Now.Millisecond);
        }
        else if( IsFlagSet(Tracer.OutputFlags.TimeTicks) )
        {
            TraceText.AppendFormat(" tick-{0}", System.DateTime.Now.Ticks);
        }

        if( IsFlagSet(Tracer.OutputFlags.PID) )
        {
            TraceText.AppendFormat(" pid-{0}", Tracer.GetCurrentProcessId());
        }

        if( IsFlagSet(Tracer.OutputFlags.TID) )
        {
            TraceText.AppendFormat(" tid-{0}", Tracer.GetCurrentThreadId());
        }

        if( IsFlagSet(Tracer.OutputFlags.Domain) )
        {
            TraceText.AppendFormat(" ad-{0}", Tracer.GetCurrentAppDomainName());
        }

        if( IsFlagSet(Tracer.OutputFlags.Context) )
        {
            TraceText.AppendFormat(" ctx-{0}", Tracer.GetCurrentContextId());
        }

        if( IsFlagSet(Tracer.OutputFlags.SinkSite) )
        {
            TraceText.AppendFormat(" {0}", (bClientSide ? "at-client" : "at-object"));
        }

        if( IsFlagSet(Tracer.OutputFlags.CallNature) )
        {
            TraceText.AppendFormat(" async-{0}", bAsync.ToString());
        }

        TraceText.AppendFormat( "]\r\n  {0}.{1}(",
                                msg.TypeName, msg.MethodName );

        Object[] args = (Object[])msg.MethodSignature;

        if( args.Length > 0 )
        {
            for( int n = 0; n < args.Length; n++ )
            {
                TraceText.AppendFormat( "{0}{1} {2} ({3})",
                                        (n == 0 ? " " : ", "),
                                        args[n].ToString(),
                                        msg.GetArgName(n),
                                        msg.GetArg(n) );
            }

            TraceText.Append(" )");
        }
        else
        {
            TraceText.Append(" void )");
        }

        TraceText.AppendFormat("\r\n  => {0}", msg.ReturnValue);

        OutStream.WriteLine(TraceText);
        OutStream.Flush();
    }
}

public class Tracer : IDynamicProperty, IContributeDynamicSink
{
    public enum OutputFlags
    {
        // Items to include in the output:
        //
        MethodEntry = 0x0001,
        MethodExit = 0x0002,
        Description = 0x0004,
        PID = 0x0008,
        TID = 0x0010,
        Domain = 0x0020,
        Context = 0x0040,
        SinkSite = 0x0080,
        CallNature = 0x0100,
        Date = 0x0200,
        Time = 0x0400,
        TimeMS = 0x0800,
        TimeTicks = 0x1000,
        
        // Some shortcuts:
        MethodBoth = MethodEntry | MethodExit,
        MethodContext = PID | TID | Domain | Context,
        Default = MethodBoth | Description | TID | Domain | Context | TimeMS | SinkSite,
        All = MethodBoth | Description | MethodContext | SinkSite |
              CallNature | Date | Time
    };

    private Tracer( String Note )
        : this(Note, OutputFlags.Default, System.Console.OpenStandardOutput())
    {
    }

    private Tracer( String Note, OutputFlags Flags )
        : this(Note, Flags, System.Console.OpenStandardOutput())
    {
    }

    private Tracer( String Note, OutputFlags Flags, String FileName )
        : this(Note, Flags, new FileStream(FileName, FileMode.Create))
    {
    }

    private Tracer( String Note, OutputFlags Flags, Stream OutStream )
    {
        Description = Note;
        this.Flags = Flags;
        Writer = new StreamWriter(OutStream);
    }

    static public void Wrap( ContextBoundObject obj, String Description )
    {
        Context.RegisterDynamicProperty(new Tracer(Description), obj, null);
    }

    static public void Wrap( ContextBoundObject obj,
                             String Description,
                             OutputFlags Flags )
    {
        Context.RegisterDynamicProperty(new Tracer(Description, Flags), obj, null);
    }

    static public void Wrap( ContextBoundObject obj,
                             String Description,
                             String FileName )
    {
        Context.RegisterDynamicProperty(new Tracer(Description, OutputFlags.Default, FileName), obj, null);
    }

    static public void Wrap( ContextBoundObject obj,
                             String Description,
                             Stream OutStream )
    {
        Context.RegisterDynamicProperty(new Tracer(Description, OutputFlags.Default, OutStream), obj, null);
    }

    static public void Wrap( ContextBoundObject obj,
                             String Description,
                             String FileName,
                             OutputFlags Flags )
    {
        Context.RegisterDynamicProperty(new Tracer(Description, Flags, FileName), obj, null);
    }

    static public void Wrap( ContextBoundObject obj,
                             String Description,
                             Stream OutStream,
                             OutputFlags Flags )
    {
        Context.RegisterDynamicProperty(new Tracer(Description, Flags, OutStream), obj, null);
    }

    static public void Unwrap( ContextBoundObject obj )
    {
        Context.UnregisterDynamicProperty("Tracer", obj, null);
    }

    static public int GetCurrentProcessId()
    {
       return System.Diagnostics.Process.GetCurrentProcess().Id;
    }

    static public String GetCurrentAppDomainName()
    {
        return System.AppDomain.CurrentDomain.FriendlyName;
    }

    static public int GetCurrentThreadId()
    {
        return System.Threading.Thread.CurrentThread.GetHashCode();
    }

    static public int GetCurrentContextId()
    {
        return System.Threading.Thread.CurrentContext.ContextID;
    }

    // IContributeDynamicSink
    public IDynamicMessageSink GetDynamicSink()
    {
        return new TracerSink(Description, Flags, Writer);
    }

    // IDynamicProperty
    public System.String Name
    {
        get {
            return("Tracer");
        }
    }

    private String          Description;
    private OutputFlags     Flags;
    private StreamWriter    Writer;
}
